@@ -1,17 +1,16 @@
 from __future__ import annotations
 
-import logging
+import os
 from collections import Counter
+from functools import partial
 
 import stanza
 from stanza import Pipeline
-from tqdm import trange
+from tqdm.contrib.concurrent import thread_map
 
 from plagdef.model import stopwords
 from plagdef.model.models import Document, Sentence, Word
 
-log = logging.getLogger(__name__)
-
 PRCS = 'tokenize,mwt,pos,lemma'
 PIPE_LVL = 'WARN'
 LOAD_LVL = 'INFO'
@@ -24,20 +23,15 @@ class Preprocessor:
         self._rem_stop_words = rem_stop_words
 
     def preprocess(self, lang: str, docs: list[Document], common_docs: list[Document] = None):
-        log.info('Preprocessing documents...')
         nlp_model = _nlp_pipe(lang)
         stop_words = stopwords.ENGLISH if lang == 'eng' else stopwords.GERMAN
-        # Preprocess common docs
-        parsed_common_docs = (nlp_model(common_doc.text) for common_doc in common_docs) if common_docs else ()
-        for idx, parsed_doc in enumerate(parsed_common_docs):
-            self._preprocess(common_docs[idx], parsed_doc.sentences, [], stop_words)
-        common_sent_words = [sent.words for doc_sents in (doc.sents for doc in common_docs)
-                             for sent in doc_sents] if common_docs else []
-        for idx in trange(len(docs)):
-            parsed_doc = nlp_model(docs[idx].text)
-            self._preprocess(docs[idx], parsed_doc.sentences, common_sent_words, stop_words)
+        common_word_lists = _common_word_lists(nlp_model, common_docs) if common_docs else []
+        thread_map(partial(self._preprocess, nlp_model=nlp_model,
+                           common_word_lists=common_word_lists, stop_words=stop_words),
+                   docs, max_workers=os.cpu_count(), desc='Preprocessing', unit='doc')
 
-    def _preprocess(self, doc: Document, sents, common_sent_words: list[list[Word]], stop_words: set[str]):
+    def _preprocess(self, doc: Document, nlp_model: Pipeline, common_word_lists: list[list[str]], stop_words: set[str]):
+        sents = nlp_model(doc.text).sentences
         for sent_idx, sent in enumerate(sents):
             non_punct_words = [word for word in sent.words if not word.upos == 'PUNCT']
             if self._rem_stop_words:
@@ -48,28 +42,47 @@ class Preprocessor:
                 lemma_count = Counter(sent_lemmas)
                 sentence = Sentence(sent.tokens[0].start_char, sent.tokens[-1].end_char, lemma_count, doc)
                 sentence.words = _to_words(sent.tokens, sentence)
-                if not _sent_contains_words_of_common_sent(sentence.words, common_sent_words):
-                    doc.sents.add(sentence)
+                doc.add_sent(sentence)
+                if _sent_contains_common_words(sentence.words, common_word_lists):
+                    sentence.common = True
+                else:
                     for lemma in lemma_count.keys():
                         doc.vocab[lemma] += 1
         self._join_small_sentences(doc)
+        self._remove_small_sentences(doc)
 
     def _join_small_sentences(self, doc: Document):
-        idx, sent_count = 0, len(doc.sents)
+        sents = doc.sents(include_common=True)
+        idx, sent_count = 0, len(sents)
         while idx < sent_count - 1:
-            sent1, sent2 = doc.sents[idx], doc.sents[idx + 1]
-            if sum(sent1.bow.values()) < self._min_sent_len \
-                or (sent2 == doc.sents[-1] and sum(sent2.bow.values()) < self._min_sent_len):
+            sent1, sent2 = sents[idx], sents[idx + 1]
+            if (not sent1.common and not sent2.common) and \
+                (len(sent1.words) < self._min_sent_len or
+                 (sent2 == sents[-1] and len(sent2.words) < self._min_sent_len)):
                 for lemma in sent1.bow.keys():
                     if lemma in sent2.bow:
                         doc.vocab[lemma] -= 1
                 joined_sent = Sentence(sent1.start_char, sent2.end_char, sent1.bow + sent2.bow, doc)
                 joined_sent.words = sent1.words + sent2.words
-                doc.sents.remove(sent1), doc.sents.remove(sent2)
-                doc.sents.add(joined_sent)
+                doc.remove_sent(sent1), doc.remove_sent(sent2)
+                doc.add_sent(joined_sent)
+                idx -= 1
                 sent_count -= 1
             idx += 1
 
+    def _remove_small_sentences(self, doc: Document):
+        sents = list(doc.sents())
+        idx, sent_count = 0, len(sents)
+        while idx < sent_count:
+            if len(sents[idx].words) < self._min_sent_len:
+                for lemma in sents[idx].bow.keys():
+                    doc.vocab[lemma] -= 1
+                doc.remove_sent(sents[idx]), sents.remove(sents[idx])
+                idx -= 1
+                sent_count -= 1
+            idx += 1
+        doc.vocab += Counter()  # Remove zero counts
+
 
 def _nlp_pipe(lang: str) -> Pipeline:
     if lang in LANG_CODES:
@@ -82,6 +95,18 @@ def _nlp_pipe(lang: str) -> Pipeline:
         raise UnsupportedLanguageError(f'The language "{lang}" is currently not supported.')
 
 
+def _common_word_lists(pipe: Pipeline, common_docs: list[Document]) -> list[list[str]]:
+    common_word_lists = []
+    for doc in common_docs:
+        for line in doc.text.splitlines():
+            parsed_line = pipe(line)
+            line_words = []
+            for sent in parsed_line.sentences:
+                [line_words.append(word.text.lower()) for word in filter(lambda w: not w.upos == 'PUNCT', sent.words)]
+            common_word_lists.append(line_words)
+    return common_word_lists
+
+
 def _to_words(stanza_tokens, sentence: Sentence) -> list[Word]:
     words = []
     for stanza_token in stanza_tokens:
@@ -91,10 +116,10 @@ def _to_words(stanza_tokens, sentence: Sentence) -> list[Word]:
     return words
 
 
-def _sent_contains_words_of_common_sent(sent_words: list[Word], common_sent_words: list[list[Word]]) -> bool:
-    sent_word_texts = [word.text for word in sent_words]
-    for sent_words in common_sent_words:
-        if all(word.text in sent_word_texts for word in sent_words):
+def _sent_contains_common_words(sent_words: list[Word], common_word_lists: list[list[str]]) -> bool:
+    sent_word_texts = [word.text.lower() for word in sent_words]
+    for common_word_list in common_word_lists:
+        if all(common_word in sent_word_texts for common_word in common_word_list):
             return True
     return False
 
