@@ -22,7 +22,7 @@ def generate_text_report(matches: list[DocumentPairMatches]) -> str:
     for doc_pair_matches in matches:
         doc1, doc2 = doc_pair_matches.doc_pair
         report = f'Pair({doc1.name}, {doc2.name}):\n'
-        for match in doc_pair_matches.list():
+        for match in sorted(doc_pair_matches.list(), key=lambda m: m.frag_from_doc(doc1).start_char):
             frag1, frag2 = match.frag_from_doc(doc1), match.frag_from_doc(doc2)
             report += f'  Match(Fragment({frag1.start_char}, {frag1.end_char}), Fragment({frag2.start_char},' \
                       f' {frag2.end_char}))\n'
@@ -40,7 +40,7 @@ def generate_xml_reports(matches: list[DocumentPairMatches]) -> list[DocumentPai
     for doc_pair_matches in matches:
         doc1, doc2 = doc_pair_matches.doc_pair
         root = E.report(doc1=doc1.name, doc2=doc2.name)
-        for match in doc_pair_matches.list():
+        for match in sorted(doc_pair_matches.list(), key=lambda m: m.frag_from_doc(doc1).start_char):
             frag1, frag2 = match.frag_from_doc(doc1), match.frag_from_doc(doc2)
             root.append(E.match(
                 doc1_start_char=str(frag1.start_char), doc1_end_char=str(frag1.end_char),
